// TTY-Grin
// Copyright (C) 2001-2003 Daniel Beer
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <stdio.h>
#include <unistd.h>
#include <time.h>
#include <fstream>
#include <algorithm>
#include "maildir.h"

maildir_message::maildir_message(maildir *m, const char *f) : filename(f) {
  parent=m;

  unsigned int begin, end;
  for(begin=filename.size()-1;f>=0&&f[begin]!='/';begin--);
  begin++;
  for(end=begin;f[end]&&f[end]!=':';end++);

  basename=string(f, begin, end-begin);

  unsigned char flags=0;

  if(f[end++]&&f[end++]=='2'&&f[end++]==',') {
    while(f[end]) {
      switch(f[end]) {
      case 'R': flags|=message_descriptor::ANSWERED; break;
      case 'S': flags|=message_descriptor::READ; break;
      case 'T': flags|=message_descriptor::DELETED; break;
      case 'F': flags|=message_descriptor::MARKED; break;
      }
      end++;
    }
    message_descriptor::set_flags(flags);
  }

  ifstream in(f);

  if(in.good()) {
    char buf[1024], *x;

    in.getline(buf, sizeof(buf));

    while((!(in.eof()||in.fail()))&&buf[0]&&buf[0]!='\r') {
      for(x=buf+strlen(buf)-1;x>=buf&&*x=='\r';x--);
      x[1]=0;

      if(!strncasecmp(buf, "From: ", 6)) set_from(buf+6);
      else if(!strncasecmp(buf, "Date: ", 6)) set_date(buf+6);
      else if(!strncasecmp(buf, "Subject: ", 9)) set_subject(buf+9);

      in.getline(buf, sizeof(buf));
    }

    in.close();
  }
}

maildir_message::~maildir_message(void) {
}

bool
maildir_message::extract(ostream& out) const {
  ifstream in(filename.c_str());
  if(in.bad()) return false;

  int c;

  while((c=in.get())>=0) out<<(char)c;
  in.close();

  return true;
}

void
maildir_message::set_flags(unsigned char f) {
  if(f!=get_flags()) {
    string flstr=":2,";
    if(f&message_descriptor::MARKED) flstr+='F';
    if(f&message_descriptor::ANSWERED) flstr+='R';
    if(f&message_descriptor::READ) flstr+='S';
    if(f&message_descriptor::DELETED) flstr+='T';

    string newname=parent->get_path();
    newname+="/cur/";
    newname+=basename;
    if(flstr.size()>3) newname+=flstr;

    if(!rename(filename.c_str(), newname.c_str()))
      filename=newname;

    message_descriptor::set_flags(f);
  }
}

maildir::maildir(const char *p) : path(p) {
  string dname=path;
  dname+="/new";
  scan(dname.c_str());
  dname=path;
  dname+="/cur";
  scan(dname.c_str());
}

maildir::~maildir(void) {
}

message_descriptor&
maildir::operator[](int which) {
  return contents[which];
}

int
maildir::size(void) const {
  return contents.size();
}

bool
maildir::append(const message_descriptor& msg) {
  char buf[128];

  generate_name(buf, sizeof(buf));

  string full=get_path();
  full+="/tmp/";
  full+=buf;

  ofstream out(full.c_str());

  if(out.bad()) return false;
  msg.extract(out);
  out.close();

  contents.push_back(maildir_message(this, full.c_str()));
  contents.back().message_descriptor::set_flags(~msg.get_flags());
  contents.back().set_flags(msg.get_flags());

  return true;
}

bool
maildir::append(const message& msg, unsigned char flags) {
  char buf[128];

  generate_name(buf, sizeof(buf));

  string full=get_path();
  full+="/tmp/";
  full+=buf;

  ofstream out(full.c_str());
  if(out.bad()) return false;
  msg.send_original(out, "\n");
  out.close();

  contents.push_back(maildir_message(this, full.c_str()));
  contents.back().message_descriptor::set_flags(~flags);
  contents.back().set_flags(flags);

  return true;
}

void
maildir::expunge(void) {
  for(unsigned int i=0;i<contents.size();i++)
    if(contents[i].get_flags()&message_descriptor::DELETED)
      unlink(contents[i].get_filename());
}

void
maildir::scan(const char *dir) {
  DIR *d=opendir(dir);

  if(!d) return;

  struct dirent *ent;
  vector<string> prelim;

  while((ent=readdir(d))) if(*ent->d_name!='.') {
    string full=dir;
    full+='/';
    full+=ent->d_name;
    struct stat st;
    if(!stat(full.c_str(), &st)&&st.st_mode&S_IFREG)
      prelim.push_back(full);
  }
  closedir(d);

  sort(prelim.begin(), prelim.end());

  for(unsigned int i=0;i<prelim.size();i++)
    contents.push_back(maildir_message(this, prelim[i].c_str()));
}

void
maildir::generate_name(char *buf, size_t max) {
  static int sequence=0;
  char hostname[128];

  hostname[0]=0;
  gethostname(hostname, sizeof(hostname));

  snprintf(buf, max, "%lu.%d_%d.%s",
	   time(0), getpid(), sequence++, hostname);
}
