// TTY-Grin
// Copyright (C) 2001-2003 Daniel Beer
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <errno.h>
#include <unistd.h>
#include <stdlib.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <netinet/in.h>
#include <netdb.h>
#include <unistd.h>
#include <fstream>
#include "grin.h"
#include "localmailbox.h"
#include "ctree.h"
#include "tempfile.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

using namespace std;

class tty_grin_compose_t : public ctree {
public:
  tty_grin_compose_t(message&);

protected:
  virtual void winch(void);
  virtual void draw(int, int, int, int);
  virtual bool keypress(int);

private:
  static char *headers[];
  message& msg;
  message_part *body;

  enum pgp_status_t {
    CLEAR=0, SIGNED=1, ENCRYPTED=2, BOTH=3
  } pgp_status;

  void resize(void);
  void index(message_part&, ctree_node * =0);
  void reindex(void);
  string try_send(message&);
  string pgpify(message&);

  bool command_pick_address(const char *);
  bool command_modify(const char *);
  bool command_attach(const char *);
  bool command_delete(const char *);
  bool command_edit(const char *);
  bool command_postpone(const char *);
  bool command_send(const char *);
  bool command_pgp_toggle(const char *);
  bool command_preview(const char *);
  bool command_quit(const char *);
  bool command_help(const char *);
};

void
tty_grin_compose_t::reindex(void) {
  while(get_first()) delete get_first();
  body=0;
  index(msg.root());
}

tty_grin_compose_t::tty_grin_compose_t(message& m) : ctree(3), msg(m) {
  body=0;
  pgp_status=CLEAR;
  resize();
  index(m.root());
}

char *tty_grin_compose_t::headers[]={
  N_("To"), N_("Cc"), N_("Bcc"), N_("Newsgroups"), N_("Subject"),
  N_("References"),
  0
};

void
tty_grin_compose_t::index(message_part& m, ctree_node *n) {
  ctree_node *p=new ctree_node(*this, 0, n);

  message_part::header_list::const_iterator h;

  h=m.headers().find("content-type.name");
  if(h==m.headers().end())
    h=m.headers().find("content-disposition.filename");
  if(h!=m.headers().end()) p->set_text(0, h->second.c_str());

  p->set_data((void *)&m);

  h=m.headers().find("content-type");
  p->set_text(1, h==m.headers().end()?"text/plain":h->second.c_str());
  if(m.children().empty()&&!body) body=&m;

  char buf[16];
  unsigned int size=m.content().size();
  if(size<1024) snprintf(buf, sizeof(buf), _("%d bytes"), size);
  else snprintf(buf, sizeof(buf), _("%.01f kbytes"), ((float)size)/1024.0);
  p->set_text(2, buf);

  for(message_part::child_list::iterator l=m.children().begin();
      l!=m.children().end();++l)
    index(*l, p);

  if(p->get_first()) p->set_expanded(true);
}

void
tty_grin_compose_t::winch(void) {
  SLtt_get_screen_size();
  resize();
  ctree::winch();
}

void
tty_grin_compose_t::draw(int r, int c, int rows, int cols) {
  SLsmg_set_color(SLC_MAIN);
  SLsmg_fill_region(r, c, rows, cols, ' ');

  int i;
  for(i=0;headers[i];i++) {
    SLsmg_set_color(SLC_HEADERS);
    SLsmg_gotorc(r+i+5, 0);
    SLsmg_printf("%s:", gettext(headers[i]));

    message_part::header_list::const_iterator
      h=msg.root().headers().find(headers[i]);
    if(h!=msg.root().headers().end()) {
       SLsmg_set_color(SLC_MAIN);
       SLsmg_gotorc(r+i+5, 20);
       SLsmg_write_string((char *)h->second.c_str());
    }
  }

  SLsmg_set_color(SLC_DIVIDER);
  SLsmg_fill_region(r+4, 0, 1, cols, '-');
  SLsmg_fill_region(r+i+5, 0, 1, cols, '-');

  SLsmg_set_color(SLC_STATUS);
  SLsmg_fill_region(r+rows-2, c, 1, cols, '-');
  SLsmg_gotorc(r+rows-2, c+4);
  SLsmg_printf(_(" %d bytes in body "), body->content().size());

  switch(pgp_status) {
  case ENCRYPTED: SLsmg_write_string(_("-- PGP encrypted ")); break;
  case SIGNED:    SLsmg_write_string(_("-- PGP signed ")); break;
  case BOTH:      SLsmg_write_string(_("-- PGP signed and encrypted ")); break;
  default: break;
  }

  SLsmg_set_color(SLC_MAIN);
  int line=i+6, ch=0, len;
  while(line<rows-2&&ch<(signed)body->content().size()) {
    for(len=0;len<cols&&body->content()[ch+len]!='\n';len++);
    SLsmg_gotorc(r+line, 0);
    for(i=0;i<len;i++) SLsmg_write_char(body->content()[ch+i]);
    ch+=len;
    if(body->content()[ch]=='\n') ch++;
    line++;
  }

  ctree::draw(r, c, 4, cols);
}

bool
tty_grin_compose_t::command_attach(const char *cmd) {
  char buf[1024];

  if(read_line(-1, 0, -1, _("Filename: "), buf, sizeof(buf), FILENAME)) {
    ifstream in(buf);
    
    if(in.bad()) report_error(strerror(errno));
    else {
      const char *base;
      message_part att;

      for(base=buf+strlen(buf)-1;*base!='/'&&base>=buf;base--);
      base++;
      
      int c;
      while((c=in.get())>=0) att.content().push_back(c);
      in.close();
      
      att.headers()["Content-Type"]=config_name_to_mime(buf);
      att.headers()["Content-Type.name"]=base;
      att.headers()["Content-Type.charset"]="us-ascii";
      
      att.headers()["Content-Transfer-Encoding"]="base64";
      
      att.headers()["Content-Disposition"]="attachment";
      att.headers()["Content-Disposition.filename"]=base;
      
      message_part::header_list::const_iterator i=
	msg.root().headers().find("content-type");
      
      if(i==msg.root().headers().end()||
	 strncasecmp(i->second.c_str(), "multipart/", 10)) {
	char boundary[64];
	
	snprintf(boundary, sizeof(boundary), "TTY-Grin_%lx.%lx",
		 time(NULL), random());
	
	message_part mp;
	message_part::header_list::iterator i=msg.root().headers().begin();
	
	while(i!=msg.root().headers().end()) {
	  if(strncasecmp(i->first.c_str(), "content-", 8)) {
	    mp.headers()[i->first]=i->second;
	    msg.root().headers().erase(i++);
	  } else ++i;
	}
	
	mp.receipts()=msg.root().receipts();
	
	mp.headers()["Content-Type"]="multipart/mixed";
	mp.headers()["Content-Type.boundary"]=boundary;
	
	mp.children().push_back(msg.root());
	msg.root()=mp;
      }
      
      msg.root().children().push_back(att);
      reindex();
    }
  }

  return true;
}

bool
tty_grin_compose_t::command_delete(const char *cmd) {
  if(get_cursor()) {
    message_part *att=(message_part *)(get_cursor()->get_data());
    if(att->children().size())
      report_error(_("You must delete the contained parts first"));
    else if(att==body)
      report_error(_("You can't delete the body of the message"));
    else if(!get_cursor()->get_parent())
      report_error(_("You can't delete the only part of the message"));
    else if(yes_or_no(_("Delete this attachment (yes or no)? "))) {
      message_part *parent=
	(message_part *)(get_cursor()->get_parent()->get_data());
      message_part::child_list::iterator i=parent->children().begin();
      while(i!=parent->children().end()&&(&(*i))!=att) ++i;
      if(i!=parent->children().end()) {
	parent->children().erase(i);
	reindex();
      }
    }
  } else SLtt_beep();
  
  return true;
}

bool
tty_grin_compose_t::command_edit(const char *cmd) {
  if(body) {
    tempfile t;
    ofstream out(t.filename());
    if(out.bad()) return true;
    for(message_part::content_t::const_iterator i=body->content().begin();
	i!=body->content().end();i++)
      out<<*i;
    out.close();
    char command[1024];
    snprintf(command, sizeof(command), "%s %s",
	     config_get("editor"), t.filename());
    SLsmg_suspend_smg();
    system(command);
    SLsmg_resume_smg();
    ifstream in(t.filename());
    if(in.bad()) return true;
    body->content().clear();
    int c;
    while((c=in.get())>=0) body->content().push_back((char)c);
    in.close();
  } else report_error("Message has no body!");

  return true;
}

bool
tty_grin_compose_t::command_help(const char *cmd) {
  static tty_grin_help_key_t help_keys[]={
    {"compose", "modify to",         N_("To")},
    {"compose", "modify cc",         N_("Cc")},
    {"compose", "modify bcc",        N_("Bcc")},
    {"compose", "modify newsgroups", N_("Newsgroups")},
    {"compose", "modify subject",    N_("Subject")},
    {"compose", "modify references", N_("References")},
    {"compose", "attach",            N_("Attach")},
    {"compose", "delete",            N_("Delete attachment")},
    {"compose", "edit",              N_("Edit body")},
    {"compose", "postpone",          N_("Postpone")},
    {"compose", "send",              N_("Send")},
    {"compose", "preview",           N_("Preview")},
    {"default", "quit",              N_("Cancel")},
    {"compose", "pick-address to",   N_("To (addressbook)")},
    {"compose", "pick-address cc",   N_("Cc (addressbook)")},
    {"compose", "pick-address bcc",  N_("Bcc (addressbook)")},
    {"compose", "pgp-toggle",        N_("Toggle PGP status")},
    {0, 0, 0}
  };

  static tty_grin_help_t help={
    N_("Compose"),
    N_("Write and send news and email messsages, or postpone them to be sent "
       "at a later date.  Postponed messages can be resumed by selecting the "
       "appropriate option on the main menu."),
    help_keys
  };

  tty_grin_help(help);
  return true;
}

bool
tty_grin_compose_t::command_postpone(const char *cmd) {
  char buf[1024];
  
  if(read_line(-1, 0, -1, _("Filename: "), buf, sizeof(buf), FILENAME)&&
     (access(buf, F_OK)||
      yes_or_no(_("Overwrite existing file (yes or no)? ")))) {
    ofstream out(buf);
    
    if(out.bad()) report_error(strerror(errno));
    else {
      msg.send(out, "\n");
      out.close();
      return false;
    }
  }

  return true;
}

bool
tty_grin_compose_t::command_send(const char *cmd) {
  if(yes_or_no(_("Are you sure you want to send this "
		 "message (yes or no)? "))) {
    message m2=msg;
    string s=pgpify(m2);
    
    if(s.size()) {
      s=_("PGP error: ")+s;
      report_error(s.c_str());
    } else {
      s=try_send(m2);
      if(s.size()) {
	if(s[s.size()-1]=='\r') s.erase(s.size()-1, 1);
	report_error(s.c_str());
      } else {
	const char *outbox=config_get("outbox");
	
	if(outbox) {
	  localmailbox out(outbox);
	  out.append(msg, message_descriptor::READ);
	}
	return false;
      }
    }
  }

  return true;
}

bool
tty_grin_compose_t::command_pgp_toggle(const char *cmd) {
  pgp_status=(pgp_status_t)((pgp_status+1)%4);
  return true;
}

bool
tty_grin_compose_t::command_preview(const char *cmd) {
  tty_grin_message(msg);
  return true;
}

bool
tty_grin_compose_t::command_quit(const char *cmd) {
  return !yes_or_no(_("Abandon this message (yes or no)? "));
}

bool
tty_grin_compose_t::command_pick_address(const char *cmd) {
  char header[64];

  for(cmd+=13;*cmd&&isspace(*cmd);cmd++);
  if(!*cmd) return true;
  strncpy(header, cmd, sizeof(header));
  header[sizeof(header)-1]=0;

  for(int i=0;header[i];i++) {
    if(isspace(header[i])) header[i]=0;
    else if(!i||header[i-1]=='-') header[i]=toupper(header[i]);
    else header[i]=tolower(header[i]);
  }

  char buf[1024];

  if(tty_grin_pickaddr(buf, sizeof(buf))) {
    msg.root().headers()[header]+=buf;
    msg.root().headers()[header]+=' ';
  }

  return true;
}

bool
tty_grin_compose_t::command_modify(const char *cmd) {
  char header[64];

  for(cmd+=7;*cmd&&isspace(*cmd);cmd++);
  if(!*cmd) return true;
  strncpy(header, cmd, sizeof(header));
  header[sizeof(header)-1]=0;

  for(int i=0;header[i];i++) {
    if(isspace(header[i])) header[i]=0;
    else if(!i||header[i-1]=='-') header[i]=toupper(header[i]);
    else header[i]=tolower(header[i]);
  }

  string prompt=gettext(header);
  prompt+=": ";
  char buf[1024];
  if(read_line(-1, 0, -1, prompt.c_str(), buf, sizeof(buf))) {
    if(buf[0]) msg.root().headers()[header]=buf;
    else {
      message_part::header_list::iterator
	i=msg.root().headers().find(header);
      if(i!=msg.root().headers().end()) msg.root().headers().erase(i);
    }
  }

  return true;
}

bool
tty_grin_compose_t::keypress(int key) {
  const char *cmd=kbd_translate("compose", key);

  static struct {
    const char *command;
    bool (tty_grin_compose_t::*func)(const char *);
  } commands[]={
    {"attach",     &tty_grin_compose_t::command_attach},
    {"delete",     &tty_grin_compose_t::command_delete},
    {"edit",       &tty_grin_compose_t::command_edit},
    {"help",       &tty_grin_compose_t::command_help},
    {"postpone",   &tty_grin_compose_t::command_postpone},
    {"send",       &tty_grin_compose_t::command_send},
    {"pgp-toggle", &tty_grin_compose_t::command_pgp_toggle},
    {"preview",    &tty_grin_compose_t::command_preview},
    {"quit",       &tty_grin_compose_t::command_quit},
    {0, 0}
  };

  if(cmd) {
    int i=0;
    while(commands[i].command&&
          strcasecmp(commands[i].command, cmd)) i++;
    
    if(commands[i].command) return (this->*commands[i].func)(cmd);
    else if(!strncasecmp(cmd, "modify ", 7)) return command_modify(cmd);
    else if(!strncasecmp(cmd, "pick-address ", 13))
      return command_pick_address(cmd);
  }

  return ctree::keypress(key);
}

void
tty_grin_compose_t::resize(void) {
  int i=SLtt_Screen_Cols/3;

  set_width(0, i);
  set_width(1, i);
  set_width(2, i);
}

string
tty_grin_compose_t::try_send(message& msg) {
  struct {
    const char *name;
    string restore;
  } save[]={
    {"To", ""},
    {"Cc", ""},
    {"Bcc", ""},
    {0, ""}
  };
  bool have_recipients=false, by_mail=false;

  {
    char buf[128], hostname[128];

    if(!gethostname(hostname, sizeof(hostname))&&
       strchr(hostname, '.')) {
      snprintf(buf, sizeof(buf), "<%s-%lx-%lx@%s>",
	       getenv("USER"), time(NULL), random(), hostname);
      msg.root().headers()["Message-ID"]=buf;
    }
  }

  message_part::header_list::iterator
    n=msg.root().headers().find("newsgroups");

  if(n!=msg.root().headers().end()) {
    string oldng=n->second;

    have_recipients=true;
    char buf[n->second.size()+1];
    strncpy(buf, n->second.c_str(), n->second.size());
    buf[n->second.size()]=0;
    
    char *sit=strtok(buf, " ");
    
    while(sit) {
      char argh[128];

      char *server=strchr(sit, '@');
      if(server) *(server++)=0;
      else {
	strncpy(argh, config_get("nntp-default"), sizeof(argh));
	argh[sizeof(argh)-1]=0;
	server=argh;
      }

      char *port=strchr(server, ':');
      if(port) *(port++)=0;
      else port="119";
      
      struct hostent *ent=gethostbyname(server);
      if(!ent) return hstrerror(h_errno);

      int sock=socket(PF_INET, SOCK_STREAM, 0);
      if(sock<0) return strerror(errno);

      struct sockaddr_in peer;
      peer.sin_family=AF_INET;
      peer.sin_port=htons(atoi(port));
      peer.sin_addr=*(struct in_addr *)ent->h_addr;

      if(connect(sock, (struct sockaddr *)&peer, sizeof(peer))<0) {
	close(sock);
	return strerror(errno);
      }

      fstream nntp(sock);
      
      char response[1024];
      nntp.getline(response, sizeof(response));
      if(nntp.fail()||atoi(response)!=200) {
	shutdown(sock, 2);
	nntp.close();
	return response;
      }

      nntp<<"POST\r\n";
      nntp.flush();
      nntp.getline(response, sizeof(response));
      if(nntp.fail()||atoi(response)!=340) {
	shutdown(sock, 2);
	nntp.close();
	return response;
      }

      for(int i=0;save[i].name;i++ ){
	message_part::header_list::iterator
	  h=msg.root().headers().find(save[i].name);
	if(h!=msg.root().headers().end()) {
	  save[i].restore=h->second;
	  msg.root().headers().erase(h);
	}
      }

      n->second=sit;
      msg.send(nntp, "\r\n");

      for(int i=0;save[i].name;i++ ){
	if(save[i].restore.size())
	  msg.root().headers()[save[i].name]=save[i].restore;
      }

      nntp<<".\r\n";
      nntp.flush();
      nntp.getline(response, sizeof(response));
      if(nntp.fail()||atoi(response)!=240) {
	shutdown(sock, 2);
	nntp.close();
	return response;
      }

      nntp<<"QUIT\r\n";
      nntp.flush();
      nntp.getline(response, sizeof(response));

      shutdown(sock, 2);
      nntp.close();
      
      sit=strtok(0, " ");
    }

    n->second=oldng;
  }

  for(int i=0;save[i].name;i++)
    if(msg.root().headers().find(save[i].name)!=
       msg.root().headers().end()) by_mail=true;

  if(by_mail) {
    have_recipients=true;
    int p[2];

    if(pipe(p)<0) return strerror(errno);
    switch(fork()) {
    case -1:
      close(p[0]);
      close(p[1]);
      return strerror(errno);
    case 0:
      close(p[1]);
      dup2(p[0], 0);
      execlp("/bin/sh", "sh", "-c", config_get("sendmail"), 0);
      exit(-1);
    }

    close(p[0]);
    ofstream out(p[1]);

    for(int i=0;save[i].name;i++) {
      message_part::header_list::iterator
	h=msg.root().headers().find(save[i].name);
      
      if(h!=msg.root().headers().end()) {
	string subbed;
	
	for(unsigned int j=0;j<h->second.size();j++) {
	  if(h->second[j]=='$') {
	    string book, name;
	    bool has_name=false;
	    unsigned int k=j+1;
	    
	    while(k<h->second.size()&&
		  !isspace(h->second[k])&&h->second[k]!=':')
	      book+=h->second[k++];
	    
	    if(h->second[k]==':') {
	      has_name=true;
	      k++;
	      while(k<h->second.size()&&!isspace(h->second[k]))
		name+=h->second[k++];
	    }
	    
	    book="/.ttygrin/addresses:"+book;
	    book=getenv("HOME")+book;
	    ifstream bin(book.c_str());
	    
	    if(bin.good()) {
	      char buf[1024];
	      
	      bin.getline(buf, sizeof(buf));
	      while(!(bin.eof()||bin.fail())) {
		char *key=strtok(buf, "\t"),
		  *comment=strtok(0, "\t"),  *addr=strtok(0, "\t");
		if(key&&comment&&addr&&
		   !(has_name&&strcasecmp(key, name.c_str()))) {
		  if(subbed.size()&&!isspace(subbed[subbed.size()-1]))
		    subbed+=' ';
		  subbed+=comment;
		  subbed+=" <";
		  subbed+=addr;
		  subbed+='>';
		}
		bin.getline(buf, sizeof(buf));
	      }
	      j=k-1;
	    } else subbed+='$';
	  } else subbed+=h->second[j];
	}
	
	save[i].restore=h->second;
	h->second=subbed;
      }
    }
    
    msg.send(out, "\n");

    for(int i=0;save[i].name;i++) {
      message_part::header_list::iterator
	h=msg.root().headers().find(save[i].name);

      if(h!=msg.root().headers().end()) h->second=save[i].restore;
    }

    out.close();
  }

  return have_recipients?"":_("No recipients specified");
}

void
tty_grin_compose(message& msg) {
  tty_grin_compose_t c(msg);

  msg.no_raw();
  c.set_cursor_color(SLC_CURSOR);
  c.run();
}

string
tty_grin_compose_t::pgpify(message& msg) {
  if(pgp_status==CLEAR) return "";

  const char *command=0;
  switch(pgp_status) {
  case SIGNED:    command="pgp-sign"; break;
  case ENCRYPTED: command="pgp-encrypt"; break;
  case BOTH:      command="pgp-sign-and-encrypt"; break;
  default: break;
  }

  if(!(command=config_get(command)))
    return _("Required command missing from configuration");

  char passphrase[32];

  if((pgp_status==SIGNED||pgp_status==BOTH)&&
     !read_line(-1, 0, -1, _("PGP passphrase: "),
		passphrase, sizeof(passphrase), PASSWORD))
    return _("Passphrase required");

  // Separate content from main headers
  message_part mp;
  message_part::header_list::iterator i=msg.root().headers().begin();
  
  while(i!=msg.root().headers().end()) {
    if(strncasecmp(i->first.c_str(), "content-", 8)) {
      mp.headers()[i->first]=i->second;
      msg.root().headers().erase(i++);
    } else ++i;
  }

  const char *reterr="";

  if(pgp_status==SIGNED) {
    tempfile content, signature;
    
    ofstream out(content.filename());
    if(out.good()) {
      msg.send(out, "\n");
      out.close();

      int p[2];
      if(!pipe(p)) {
	int pid=fork();

	char buf[1024];
	const char *argv[]={content.filename(), signature.filename()};
	pgp_printf(buf, sizeof(buf), command, 2, argv);

	switch(pid) {
	case -1:
	  close(p[0]);
	  close(p[1]);
	  reterr=strerror(errno);
	  break;
	case 0:
	  dup2(p[0], 0);
	  close(p[1]);
	  execl("/bin/sh", "sh", "-c", buf, 0);
	  exit(-1);
	default:
	  close(p[0]);
	  write(p[1], passphrase, strlen(passphrase));
	  close(p[1]);
	  waitpid(pid, 0, 0);

	  ifstream in(signature.filename());
	  if(in.good()) {
	    int k=in.get();
	    if(k>=0) {
	      message_part sig;
	      while(k>=0) {
		sig.content().push_back(k);
		k=in.get();
	      }

	      snprintf(buf, sizeof(buf), "TTY-Grin_%lx.%lx",
		       time(0), random());

	      sig.headers()["Content-Type"]="application/pgp-signature";
	      sig.headers()["Content-Disposition"]="inline";
	      mp.headers()["Content-Type"]="multipart/signed";
	      mp.headers()["Content-Type.micalg"]="pgp-sha1";
	      mp.headers()["Content-Type.protocol"]=
		"application/pgp-signature";
	      mp.headers()["Content-Type.boundary"]=buf;
	      mp.children().push_back(msg.root());
	      mp.children().push_back(sig);
	      msg.root()=mp;
	    } else reterr=_("PGP returned no data");
	    in.close();
	  } else reterr=strerror(errno);
	  break;
	}
      } else reterr=strerror(errno);
    } else reterr=strerror(errno);
  } else if(pgp_status==ENCRYPTED) {
    message_part::header_list::const_iterator
      r=mp.headers().find("to");

    if(r!=mp.headers().end()) {
      tempfile content, cipher;
      
      ofstream out(content.filename());
      if(out.good()) {
	msg.send(out, "\n");
	out.close();
	
	char buf[1024];
	// Recipient name may be unsafe
	const char *argv[]={content.filename(), cipher.filename(),
			    r->second.c_str()};
	pgp_printf(buf, sizeof(buf), command, 3, argv);
	system(buf);

	ifstream in(cipher.filename());
	if(in.good()) {
	  int k=in.get();
	  if(k>=0) {
	    message_part nc;
	    while(k>=0) {
	      nc.content().push_back(k);
	      k=in.get();
	    }
	    nc.headers()["Content-Type"]="application/octet-stream";
	    nc.headers()["Content-Disposition"]="inline";

	    message_part tag;

	    tag.headers()["Content-Type"]="application/pgp-encrypted";
	    tag.headers()["Content-Disposition"]="attachment";
	    for(const char *x="Version: 1\n";*x;x++)
	      tag.content().push_back(*x);

	    mp.headers()["Content-Type"]="multipart/encrypted";
	    mp.headers()["Content-Type.protocol"]="application/pgp-encrypted";
	    snprintf(buf, sizeof(buf), "TTY-Grin_%lx.%lx",
		     time(0), random());
	    mp.headers()["Content-Type.boundary"]=buf;
	    mp.children().push_back(tag);
	    mp.children().push_back(nc);
	    msg.root()=mp;
	  } else reterr=_("PGP returned no data");
	} else reterr=strerror(errno);
      } else reterr=strerror(errno);
    } else reterr=strerror(errno);
  } else {
    message_part::header_list::const_iterator
      r=mp.headers().find("to");

    if(r!=mp.headers().end()) {
      tempfile content, cipher;
      
      ofstream out(content.filename());
      if(out.good()) {
	msg.send(out, "\n");
	out.close();
	
	char buf[1024];
	const char *argv[]={content.filename(), cipher.filename(),
			    r->second.c_str()};
	pgp_printf(buf, sizeof(buf), command, 3, argv);

	int p[2];
	if(!pipe(p)) {
	  int pid=fork();

	  switch(pid) {
	  case -1:
	    close(p[0]);
	    close(p[1]);
	    reterr=strerror(errno);
	    break;
	  case 0:
	    close(p[1]);
	    dup2(p[0], 0);
	    execl("/bin/sh", "sh", "-c", buf, 0);
	    exit(-1);
	  default:
	    close(p[0]);
	    write(p[1], passphrase, strlen(passphrase));
	    close(p[1]);
	    waitpid(pid, 0, 0);

	    ifstream in(cipher.filename());
	    if(in.good()) {
	      int k=in.get();
	      if(k>=0) {
		message_part nc;
		while(k>=0) {
		  nc.content().push_back(k);
		  k=in.get();
		}
		nc.headers()["Content-Type"]="application/octet-stream";
		nc.headers()["Content-Disposition"]="inline";
		
		message_part tag;
		
		tag.headers()["Content-Type"]="application/pgp-encrypted";
		tag.headers()["Content-Disposition"]="attachment";
		for(const char *x="Version: 1\n";*x;x++)
		  tag.content().push_back(*x);
		
		mp.headers()["Content-Type"]="multipart/encrypted";
		mp.headers()["Content-Type.protocol"]=
		  "application/pgp-encrypted";
		snprintf(buf, sizeof(buf), "TTY-Grin_%lx.%lx",
			 time(0), random());
		mp.headers()["Content-Type.boundary"]=buf;
		mp.children().push_back(tag);
		mp.children().push_back(nc);
		msg.root()=mp;
	      } else reterr=_("PGP returned no data");
	    } else reterr=strerror(errno);
	    break;
	  }
	} else reterr=strerror(errno);
      } else reterr=strerror(errno);
    } else reterr=strerror(errno);
  }

  memset(passphrase, 0, sizeof(passphrase));
  return reterr;
}
