// TTY-Grin
// Copyright (C) 2001-2003 Daniel Beer
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#define _GNU_SOURCE
#include <unistd.h>
#include <stdio.h>
#include <fnmatch.h>
#include <regex.h>
#include <pwd.h>
#include <string>
#include <map>
#include <fstream>
#include "grin.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

using namespace std;

struct lt_string {
  bool operator()(const string& a, const string& b) const { return a<b; }
};

map<string, string, lt_string> config;
list<pair<string, string> > config_mime;
list<pair<string, string> > config_handler;
list<pair<string, string> > config_filter;
list<string> config_kill;

void
config_line(char *buf) {
  char errbuf[SLtt_Screen_Cols+1];

  char *one=strtok(buf, " \t\r\n");
  if(one) {
    if(!strcasecmp(one, "set")) {
      char *option=strtok(0, " \t\r\n"),
	*value=strtok(0, "");
      if(value) {
	for(int i=0;option[i];i++) option[i]=tolower(option[i]);
	config[option]=value;
      }
    } else if(!strcasecmp(one, "mime-type")) {
      char *type=strtok(0, " \t\r\n"), *ext;
      
      if(type) {
	for(int i=0;type[i];i++) type[i]=tolower(type[i]);
	while((ext=strtok(0, " \t\r\n")))
	  config_mime.push_back(pair<string, string>(type, ext));
      }
    } else if(!strcasecmp(one, "mime-handler")) {
      char *type=strtok(0, " \t\r\n"),
	*cmd=strtok(0, "");
      
      if(cmd) config_handler.push_back(pair<string, string>(type, cmd));
    } else if(!strcasecmp(one, "mime-filter")) {
      char *type=strtok(0, " \t\r\n"),
	*cmd=strtok(0, "");

      if(cmd) config_filter.push_back(pair<string, string>(type, cmd));
    } else if(!strcasecmp(one, "kill")) {
      char *who=strtok(0, "");
      if(who) config_kill.push_back(who);
    } else if(!strcasecmp(one, "bind")) {
      char *domain=strtok(0, " \t\r\n"),
	*key=strtok(0, " \t\r\n"),
	*cmd=strtok(0, "");

      if(cmd) {
	int code=kbd_name_to_key(key);
	
	if(code==SL_KEY_ERR) {
	  snprintf(errbuf, sizeof(errbuf), "Unknown key: %s", key);
	  report_error(errbuf);
	} else kbd_bind(domain, code, cmd);
      }
    } else if(!strcasecmp(one, "color")) {
      char *name=strtok(0, " \t\r\n"),
	*fg=strtok(0, " \t\r\n"), *bg=strtok(0, " \t\r\n");
      
      if(bg) {
	static struct {
	  int attr;
	  char *name;
	} color_names[]={
	  {SLC_MAIN,      "main"},
	  {SLC_STATUS,    "status"},
	  {SLC_HEADERS,   "headers"},
	  {SLC_DIVIDER,   "divider"},
	  {SLC_CURSOR,    "cursor"},
	  {SLC_SELECTION, "selection"},
	  {SLC_QUOTED,    "quoted"},
	  {SLC_SIGNATURE, "signature"},
	  {SLC_INFO,      "info"},
	  {0, 0}
	};
	
	int i=0;
	while(color_names[i].name&&
	      strcasecmp(color_names[i].name, name)) i++;
	
	if(color_names[i].name)
	  SLtt_set_color(color_names[i].attr, color_names[i].name, fg, bg);
	else {
	  snprintf(errbuf, sizeof(errbuf), _("Unknown color: %s"), one);
	  report_error(errbuf);
	}
      }
    } else {
      snprintf(errbuf, sizeof(errbuf), _("Unknown config option: %s"), one);
      report_error(errbuf);
    }
  }
}

void
config_load(const char *filename) {
  ifstream in(filename);
  char buf[1024], line[1024];
  int end=0, len;

  in.getline(buf, sizeof(buf));
  while(!(in.fail()||in.eof())) {
    if(buf[0]!='#') {
      len=strlen(buf);
      strncpy(line+end, buf, sizeof(line)-end);
      end+=len;
      if(end>=sizeof(buf)) end=sizeof(buf)-1;
      
      if(len&&buf[len-1]=='\\') end--;
      else {
	line[end]=0;
	config_line(line);
	end=0;
      }
    }
    in.getline(buf, sizeof(buf));
  }

  in.close();
}

void
config_init(void) {
  string home=getenv("HOME");
  string user=getenv("USER");

  config["mail"]=home+"/mail";
  config["inbox"]=
    (access("/var/mail", R_OK)?"/var/spool/mail/":"/var/mail/")+user;

  char host[128];
  if(gethostname(host, sizeof(host))) {
    config["nntp-default"]="localhost";
    config["from"]=user;
  } else {
    config["nntp-default"]=host;
    config["from"]=user+"@"+host;
  }

  char *env;
  if((env=getenv("EDITOR"))) config["editor"]=env;
  else config["editor"]="vi";

  config["sendmail"]=
    access("/usr/sbin/sendmail", X_OK)?"sendmail -t":"/usr/sbin/sendmail -t";

  config["mime-type-default"]="text/plain";

  snprintf(host, sizeof(host), "%s/.signature", getenv("HOME"));
  config["signature"]=host;

  struct passwd *p=getpwuid(getuid());
  char *x=strchr(p->pw_gecos, ',');
  if(x) *x=0;
  config["real-name"]=p->pw_gecos;

  config["index-threshold"]="10240";

  snprintf(host, sizeof(host), "%s/.ttygrin/rc", getenv("HOME"));
  config_load(host);
}

const char *
config_get(const char *name) {
  map<string, string, lt_string>::const_iterator i=config.find(name);

  if(i==config.end()) return 0;
  return i->second.c_str();
}

const char *
config_name_to_mime(const char *name) {
  const char *ext;

  for(ext=name+strlen(name);ext>=name&&*ext!='.';ext--);
  ext++;

  for(list<pair<string, string> >::const_iterator i=config_mime.begin();
      i!=config_mime.end();++i)
    if(!strcasecmp(i->second.c_str(), ext)) return i->first.c_str();

  return config_get("mime-type-default");
}

const char *
config_find(list<pair<string, string> >& l, const char *type) {
  for(list<pair<string, string> >::const_iterator i=l.begin();
      i!=l.end();++i)
    if(!fnmatch(i->first.c_str(), type, FNM_CASEFOLD))
      return i->second.c_str();

  return 0;
}

const char *
config_mime_handler(const char *type) {
  return config_find(config_handler, type);
}

const char *
config_mime_filter(const char *type) {
  return config_find(config_filter, type);
}

bool
config_should_kill(const char *who) {
  for(list<string>::const_iterator i=config_kill.begin();i!=config_kill.end();
      ++i) {
    regex_t preg;

    if(!regcomp(&preg, i->c_str(), REG_EXTENDED|REG_NOSUB)&&
       !regexec(&preg, who, 0, 0, 0)) return true;
  }

  return false;
}
