// TTY-Grin
// Copyright (C) 2001-2003 Daniel Beer
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <sys/types.h>
#include <dirent.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <fstream>
#include "imap4boxlist.h"
#include "grin.h"
#include "ctree.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

using namespace std;

class tty_grin_imap4_t : public ctree {
public:
  tty_grin_imap4_t(void);
  virtual ~tty_grin_imap4_t(void);

protected:
  virtual void draw(int, int, int, int);
  virtual bool keypress(int);
  virtual void winch(void);

private:
  bool commit;

  void resize(void);

  bool command_modify(const char *);
  bool command_create(const char *);
  bool command_delete(const char *);
  bool command_select(const char *);
  bool command_help(const char *);
};

tty_grin_imap4_t::tty_grin_imap4_t(void) : ctree(4) {
  commit=false;
  int f=open(".", O_RDONLY, 0);

  if(!(chdir(getenv("HOME"))||chdir(".ttygrin"))) {
    ifstream in("imap4");

    if(in.good()) {
      char buf[1024];

      in.getline(buf, sizeof(buf));
      while(!(in.eof()||in.fail())) {
	char *x=buf;
	ctree_node *n=new ctree_node(*this);

	for(int i=0;i<4;i++) {
	  char *y=strsep(&x, "\t");
	  if(y) n->set_text(i, y);
	}
	in.getline(buf, sizeof(buf));
      }
      in.close();
    }
  }

  fchdir(f);
  close(f);

  resize();
}

tty_grin_imap4_t::~tty_grin_imap4_t(void) {
  if(commit) {
    string name=getenv("HOME");
    name+="/.ttygrin/imap4";
    ofstream out(name.c_str());
    if(out.good()) {
      for(const ctree_node *n=get_first();n;n=n->get_next()) {
	for(int i=0;i<3;i++) out<<n->get_text(i)<<'\t';
	out<<n->get_text(3)<<endl;
      }
    } else report_error(strerror(errno));
  }
}

void
tty_grin_imap4_t::winch(void) {
  SLtt_get_screen_size();
  resize();
  ctree::winch();
}

void
tty_grin_imap4_t::resize(void) {
  set_width(0, SLtt_Screen_Cols/3);
  set_width(1, SLtt_Screen_Cols/3);
  set_width(2, SLtt_Screen_Cols/3);
}

void
tty_grin_imap4_t::draw(int r, int c, int rows, int cols) {
  SLsmg_set_color(SLC_HEADERS);
  SLsmg_fill_region(r, c, 1, cols, ' ');
  SLsmg_gotorc(r, 6);
  SLsmg_write_string(_("Host"));
  SLsmg_gotorc(r, 6+get_width(0));
  SLsmg_write_string(_("Username"));
  SLsmg_gotorc(r, 6+get_width(0)+get_width(1));
  SLsmg_write_string(_("Prefix"));

  SLsmg_set_color(SLC_STATUS);
  SLsmg_fill_region(r+rows-2, c, 1, cols, '-');
  SLsmg_gotorc(r+rows-2, c+4);
  SLsmg_printf(_(" IMAP4 profiles "));

  SLsmg_set_color(SLC_MAIN);
  SLsmg_fill_region(r+rows-1, c, 1, cols, ' ');

  ctree::draw(r+1, c, rows-3, cols);
}

bool
tty_grin_imap4_t::command_modify(const char *cmd) {
  if(get_cursor()) {
    int which=0;
    switch(tolower(cmd[8])) {
    case 'o': which=0; break;
    case 's': which=1; break;
    case 'r': which=2; break;
    case 'a': which=3; break;
    }
    static char *prompt[]={N_("Host: "), N_("Username: "), N_("Prefix: "),
			   N_("Password: ")};
    char buf[128];
    if(read_line(-1, 0, -1, gettext(prompt[which]), buf, sizeof(buf),
		 which=='3'?PASSWORD:TEXT)) {
      get_cursor()->set_text(which, buf);
      commit=true;
    }
  }

  return true;
}

bool
tty_grin_imap4_t::command_create(const char *cmd) {
  char host[128];
  char username[128];
  char prefix[128];
  char password[128];
  
  if(read_line(-1, 0, -1, _("Host: "), host, sizeof(host))&&
     read_line(-1, 0, -1, _("Username: "), username, sizeof(username))&&
     read_line(-1, 0, -1, _("Prefix: "), prefix, sizeof(prefix))&&
     read_line(-1, 0, -1, _("Password: "), password, sizeof(password),
	       PASSWORD)) {
    ctree_node *n=new ctree_node(*this);
    n->set_text(0, host);
    n->set_text(1, username);
    n->set_text(2, prefix);
    n->set_text(3, password);
    commit=true;
  }

  return true;
}

bool
tty_grin_imap4_t::command_delete(const char *cmd) {
  if(get_cursor()&&yes_or_no(_("Delete this profile (yes or no)? "))) {
    delete get_cursor();
    commit=true;
  }

  return true;
}

bool
tty_grin_imap4_t::command_help(const char *cmd) {
  static tty_grin_help_key_t help_keys[]={
    {"imap4",   "modify-hostname", N_("Modify hostname")},
    {"imap4",   "modify-user",     N_("Modify user")},
    {"imap4",   "modify-prefix",   N_("Modify prefix")},
    {"imap4",   "modify-password", N_("Modify password")},
    {"imap4",   "create",          N_("Create")},
    {"imap4",   "delete",          N_("Delete")},
    {"default", "quit",            N_("Back")},
    {0, 0, 0}
  };

  static tty_grin_help_t help={
    N_("IMAP4 profiles"),
    N_("Create and manage profiles for Internet Mail Access Protocol, to "
       "use mailboxes on remote systems."),
    help_keys
  };

  tty_grin_help(help);
  return true;
}

bool
tty_grin_imap4_t::command_select(const char *cmd) {
  if(get_cursor()) {
    imap4boxlist bl(get_cursor()->get_text(0),
		    get_cursor()->get_text(1),
		    get_cursor()->get_text(2),
		    get_cursor()->get_text(3));
    tty_grin_boxlist(bl, get_cursor()->get_text(0));
  } else SLtt_beep();

  return true;
}

bool
tty_grin_imap4_t::keypress(int key) {
  const char *cmd=kbd_translate("imap4", key);

  static struct {
    const char *command;
    bool (tty_grin_imap4_t::*func)(const char *);
  } commands[]={
    {"modify-hostname", &tty_grin_imap4_t::command_modify},
    {"modify-user",     &tty_grin_imap4_t::command_modify},
    {"modify-prefix",   &tty_grin_imap4_t::command_modify},
    {"modify-password", &tty_grin_imap4_t::command_modify},
    {"create",          &tty_grin_imap4_t::command_create},
    {"delete",          &tty_grin_imap4_t::command_delete},
    {"select",          &tty_grin_imap4_t::command_select},
    {"help",            &tty_grin_imap4_t::command_help},
    {0, 0}
  };

  if(cmd) {
    int i=0;
    while(commands[i].command&&
          strcasecmp(commands[i].command, cmd)) i++;
    
    if(commands[i].command) return (this->*commands[i].func)(cmd);
  }

  return ctree::keypress(key);
}

void
tty_grin_imap4(void) {
  tty_grin_imap4_t n;

  n.set_cursor_color(SLC_CURSOR);
  n.run();
}
