// TTY-Grin
// Copyright (C) 2001-2003 Daniel Beer
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <fstream>
#include <signal.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <glob.h>
#include <stdlib.h>
#include <time.h>
#include <locale.h>
#include "localmailbox.h"
#include "grin.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

using namespace std;

void *sigdata_winch;

struct read_line_t {
  const char *prompt;
  char *buf;
  int plen, scroll, pos, len, row, col, numcol, max;
  void *sigdata;
  SLSig_Fun_Type *sig;
  read_type_t rlt;
};

void
read_line_draw(read_line_t& r) {
  int row=r.row;
  int col=r.col;
  int numcol=r.numcol;

  if(row<0) row+=SLtt_Screen_Rows;
  if(col<0) col+=SLtt_Screen_Cols;
  if(numcol<0) numcol=SLtt_Screen_Cols-col;

  if(r.pos<r.scroll) r.scroll=r.pos;
  if(r.pos-(numcol-r.plen)>=r.scroll)
    r.scroll=r.pos-(numcol-r.plen)+1;

  SLsmg_fill_region(row, col, 1, numcol, ' ');
  SLsmg_gotorc(row, col);
  SLsmg_write_nstring((char *)r.prompt, numcol);
  int n=r.len-r.scroll;
  if(n>numcol-r.plen) n=numcol-r.plen;
  SLsmg_gotorc(row, r.plen);
  if(r.rlt==PASSWORD)
    for(int i=0;i<n&&r.buf[r.scroll+i];i++) SLsmg_write_string("*");
  else SLsmg_write_nstring(r.buf+r.scroll, n);
  SLsmg_gotorc(row, col+r.plen+r.pos-r.scroll);
}

void
read_line_winch(int sig) {
  read_line_t& r=*(read_line_t *)sigdata_winch;

  sigdata_winch=r.sigdata;
  r.sig(sig);
  sigdata_winch=&r;

  read_line_draw(r);
  SLsmg_refresh();
}

bool
read_line_command_select(read_line_t& r, const char *cmd) {
  r.buf[r.len]=0;
  return false;
}

bool
read_line_command_cancel(read_line_t& r, const char *cmd) {
  return false;
}

bool
read_line_command_redraw(read_line_t& r, const char *cmd) {
  SLsmg_touch_lines(0, SLtt_Screen_Rows);
  return true;
}

bool
read_line_command_delete_back(read_line_t& r, const char *cmd) {
  if(r.pos) {
    memmove(r.buf+r.pos-1, r.buf+r.pos, r.len-r.pos);
    r.pos--;
    r.len--;
  }
  return true;
}

bool
read_line_command_delete_forward(read_line_t& r, const char *cmd) {
  if(r.pos<r.len) {
    memmove(r.buf+r.pos, r.buf+r.pos+1, r.len-r.pos-1);
    r.len--;
  }
  return true;
}

bool
read_line_command_back(read_line_t& r, const char *cmd) {
  if(r.pos) r.pos--;
  return true;
}

bool
read_line_command_forward(read_line_t& r, const char *cmd) {
  if(r.pos<r.len) r.pos++;
  return true;
}

bool
read_line_command_jump_top(read_line_t& r, const char *cmd) {
  r.pos=0;
  return true;
}

bool
read_line_command_jump_end(read_line_t& r, const char *cmd) {
  r.pos=r.len;
  return true;
}

bool
read_line_command_cut_forward(read_line_t& r, const char *cmd) {
  r.len=r.pos;
  return true;
}

bool
read_line_command_cut_back(read_line_t& r, const char *cmd) {
  memmove(r.buf, r.buf+r.pos, r.len-r.pos);
  r.len-=r.pos;
  r.pos=0;
  return true;
}

bool
read_line_command_complete(read_line_t& r, const char *cmd) {
  if(r.rlt==FILENAME) {
    glob_t g;
    char pattern[r.len+2];
    
    memcpy(pattern, r.buf, r.pos);
    pattern[r.pos]='*';
    memcpy(pattern+r.pos+1, r.buf+r.pos, r.len-r.pos);
    pattern[r.len+1]=0;
    
    if(glob(pattern, GLOB_NOSORT|GLOB_TILDE|GLOB_NOESCAPE|
	    GLOB_MARK, 0, &g)<0) SLtt_beep();
    else if(g.gl_pathc==1) {
      r.len=strlen(*g.gl_pathv);
      if((unsigned)r.len>=r.max) r.len=r.max-1;
      memcpy(r.buf, *g.gl_pathv, r.len);
      r.pos=r.len;
    }
    
    globfree(&g);
  } else SLtt_beep();
  return true;
}

bool
read_line(int row, int col, int numcol,
	  const char *prompt, char *buf, size_t max, read_type_t rlt) {
  read_line_t r;

  r.prompt=prompt;
  r.plen=strlen(prompt);
  r.buf=buf;
  r.scroll=0;
  r.pos=0;
  r.len=0;
  r.max=max;
  r.row=row;
  r.col=col;
  r.numcol=numcol;
  r.sigdata=sigdata_winch;
  r.rlt=rlt;
  r.sig=SLsignal(SIGWINCH, read_line_winch);
  sigdata_winch=&r;

  for(;;) {
    read_line_draw(r);
    SLsmg_refresh();

    int key=SLkp_getkey();

    if(key>=32&&key<=255&&key!=127) {
      if((unsigned)(r.len+1)<max) {
	memmove(r.buf+r.pos+1, r.buf+r.pos, r.len-r.pos);
	r.buf[r.pos++]=key;
	r.len++;
      } else SLtt_beep();
    } else {
      const char *cmd=kbd_translate("read-line", key);

      if(cmd) {
	static struct {
	  const char *command;
	  bool (*func)(read_line_t&, const char *);
	} commands[]={
	  {"select",         read_line_command_select},
	  {"cancel",         read_line_command_cancel},
	  {"redraw",         read_line_command_redraw},
	  {"delete-back",    read_line_command_delete_back},
	  {"delete-forward", read_line_command_delete_forward},
	  {"back",           read_line_command_back},
	  {"forward",        read_line_command_forward},
	  {"jump-top",       read_line_command_jump_top},
	  {"jump-end",       read_line_command_jump_end},
	  {"cut-forward",    read_line_command_cut_forward},
	  {"cut-back",       read_line_command_cut_back},
	  {"complete",       read_line_command_complete},
	  {0, 0}
	};

	int i=0;

	while(commands[i].command&&
	      strcasecmp(commands[i].command, cmd)) i++;

	if(commands[i].command) {
	  if(!commands[i].func(r, cmd)) {
	    sigdata_winch=r.sigdata;
	    SLsignal(SIGWINCH, r.sig);
	    return tolower(*cmd)=='s';
	  }
	} else SLtt_beep();
      } else SLtt_beep();
    }
  }
}

void
report_error(const char *message) {
  SLsmg_fill_region(SLtt_Screen_Rows-1, 0, 1, SLtt_Screen_Cols, ' ');
  SLsmg_gotorc(SLtt_Screen_Rows-1, 0);
  SLsmg_write_string((char *)message);
  SLsmg_refresh();
  SLkp_getkey();
}

bool
yes_or_no(const char *prompt) {
  char buf[16];

  for(;;) {
    if(read_line(-1, 0, -1, prompt, buf, sizeof(buf))) {
      if(!strcasecmp(buf, _("yes"))) return true;
      if(!strcasecmp(buf, _("no"))) return false;
    }
    report_error(_("Please answer \"yes\" or \"no\"."));
  }
}

void
check_directory(const char *path) {
  char buf[128];

  if(access(path, R_OK)) {
    if(mkdir(path, 0700))
      snprintf(buf, sizeof(buf), "%s: %s", path, strerror(errno));
    else snprintf(buf, sizeof(buf), _("Created directory %s"), path);
    report_error(buf);
  }
}

void
prepare_message(message_part& head, message_part& body) {
  const char *from=config_get("from");
  const char *real=config_get("real-name");

  if(real) {
    char buf[128];
    snprintf(buf, sizeof(buf), "%s <%s>", real, from);
    head.headers()["From"]=buf;
  } else head.headers()["From"]=from;

  if((real=config_get("reply-to"))) head.headers()["Reply-To"]=real;
  head.headers()["Content-Type"]="text/plain";
  head.headers()["Content-Disposition"]="inline";
  head.headers()["MIME-Version"]="1.0";
  head.headers()["User-Agent"]="TTY-Grin version " VERSION
    " -- Copyright (C) 2001-2003 Daniel Beer";
  if((real=config_get("organization"))) head.headers()["Organization"]=real;

  ifstream in(config_get("signature"));
  if(in.good()) {
    int i;
    body.content().push_back('\n');
    body.content().push_back('-');
    body.content().push_back('-');
    body.content().push_back(' ');
    body.content().push_back('\n');
    while((i=in.get())>=0) body.content().push_back(i);
    in.close();
  }
}

void
reaper(int sig) {
  waitpid(-1, 0, WNOHANG);
}

void
pgp_printf(char *out, size_t max, const char *format,
	   int argc, const char **argv) {
  int end=0;

  while(*format&&end+1<max) {
    if(*format=='%') {
      if(isdigit(*(++format))) {
	int which=*format-'1';
	if(which>=0&&which<argc) {
	  const char *x=argv[which];
	  while(*x&&end+1<max) out[end++]=*(x++);
	}
      }
    } else out[end++]=*format;
    format++;
  }

  out[end]=0;
}

int
main(int argc, char *argv[]) {
  signal(SIGINT, SIG_IGN);
  signal(SIGABRT, SIG_IGN);
  signal(SIGTSTP, SIG_IGN);
  signal(SIGCHLD, reaper);

  srandom(time(NULL));
  umask(077);

  setlocale(LC_ALL, "");
  bindtextdomain(PACKAGE, LOCALEDIR);
  textdomain(PACKAGE);

  SLtt_get_terminfo();
  SLkp_init();
  SLang_init_tty(-1, 0, 0);
  SLsmg_init_smg();

  SLtt_set_color(SLC_MAIN,      "main",      "lightgray", "black");
  SLtt_set_color(SLC_STATUS,    "status",    "black",     "lightgray");
  SLtt_set_color(SLC_HEADERS,   "headers",   "white",     "black");
  SLtt_set_color(SLC_DIVIDER,   "divider",   "white",     "black");
  SLtt_set_color(SLC_CURSOR,    "cursor",    "white",     "black");
  SLtt_set_color(SLC_SELECTION, "selection", "black",     "lightgray");
  SLtt_set_color(SLC_QUOTED,    "quoted",    "white",     "black");
  SLtt_set_color(SLC_SIGNATURE, "signature", "white",     "black");
  SLtt_set_color(SLC_INFO,      "info",      "white",     "black");

  kbd_init();
  config_init();
  localmailbox::set_index_threshold(atoi(config_get("index-threshold")));

  {
    string name=getenv("HOME");
    name+="/.ttygrin";
    check_directory(name.c_str());
    check_directory(config_get("mail"));
  }

  const char *to=0, *subject=0, *resume=0;

  for(int i=1;i<argc;i++) {
    if(!strcmp(argv[i], "-nocolor")) SLtt_Use_Ansi_Colors=0;
    else if(!strcmp(argv[i], "-subject")) subject=argv[++i];
    else if(!strcmp(argv[i], "-resume")) resume=argv[++i];
    else if(!strcmp(argv[i], "-version")) {
      SLsmg_reset_smg();
      SLang_reset_tty();
      cerr<<"TTY-Grin version " VERSION<<endl;
      cerr<<"Copyright (C) 2001-2003 Daniel Beer"<<endl;
      return 0;
    } else if(!strcmp(argv[i], "-bindings")) {
      SLsmg_reset_smg();
      SLang_reset_tty();
      kbd_dump(cerr);
      return 0;
    } else {
      if(to) {
	SLsmg_reset_smg();
	SLang_reset_tty();
	cerr<<"Unknown option: "<<argv[i]<<endl;
	return -1;
      } else to=argv[i];
    }
  }

  if(resume) {
    ifstream in(resume);

    if(in.bad()) report_error(strerror(errno));
    else {
      message m(in);
      in.close();
      tty_grin_compose(m);
    }
  } else if(to) {
    message m;

    prepare_message(m.root(), m.root());
    m.root().headers()["To"]=to;
    if(subject) m.root().headers()["Subject"]=subject;
    tty_grin_compose(m);    
  } else tty_grin_menu();

  SLsmg_reset_smg();
  SLang_reset_tty();

  return 0;
}
