// TTY-Grin
// Copyright (C) 2001-2003 Daniel Beer
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <signal.h>
#include <string.h>
#include <errno.h>
#include <fstream>
#include "grin.h"
#include "localmailbox.h"
#include "localboxlist.h"
#include "maildir.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

using namespace std;

struct tty_grin_menu_item_t {
  const char *name;
  const char *description;
  const char *domain;
  const char *command;
  bool (*func)(const char *);
};

struct tty_grin_menu_t {
  int count, selected;
  tty_grin_menu_item_t *items;
};

void
tty_grin_menu_draw(const tty_grin_menu_t& menu) {
  SLsmg_set_color(SLC_MAIN);
  SLsmg_fill_region(0, 0, SLtt_Screen_Rows, SLtt_Screen_Cols, ' ');

  SLsmg_set_color(SLC_STATUS);
  SLsmg_fill_region(SLtt_Screen_Rows-2, 0, 1, SLtt_Screen_Cols, '-');
  SLsmg_gotorc(SLtt_Screen_Rows-2, 4);
  SLsmg_printf(_(" TTY-Grin version %s -- Copyright (C) 2001-2003 "
		 "Daniel Beer "), VERSION);

  SLsmg_set_color(SLC_MAIN);

  for(int i=0;i<menu.count;i++) {
    if(i==menu.selected) {
      SLsmg_set_color(SLC_SELECTION);
      SLsmg_fill_region(i*2+2, 10, 1, SLtt_Screen_Cols-20, ' ');
    }
    SLsmg_gotorc(i*2+2, 11);
    string kname=kbd_reverse_translate(menu.items[i].domain,
				       menu.items[i].command);
    SLsmg_printf("%s - %s", gettext(kname.c_str()),
		 gettext(menu.items[i].name));
    SLsmg_gotorc(i*2+2, SLtt_Screen_Cols/2-10);
    SLsmg_write_string((char *)gettext(menu.items[i].description));
    if(i==menu.selected) SLsmg_set_color(SLC_MAIN);
  }
}

void
tty_grin_menu_sigwinch(int s) {
  SLtt_get_screen_size();
  SLsmg_reinit_smg();
  tty_grin_menu_draw(*(tty_grin_menu_t *)sigdata_winch);
  SLsmg_refresh();
}

bool
tty_grin_menu_command_inbox(const char *cmd) {
  const char *md=config_get("maildir");
  
  if(md) {
    maildir m(md);
    tty_grin_mailbox(m, _("Inbox"));
  } else {
    localmailbox m(config_get("inbox"));
    tty_grin_mailbox(m, _("Inbox"));
  }

  return true;
}

bool
tty_grin_menu_command_mail(const char *cmd) {
  localboxlist b(config_get("mail"));
  tty_grin_boxlist(b, _("Local mail"));
  return true;
}

bool
tty_grin_menu_command_news(const char *cmd) {
  tty_grin_news();
  return true;
}

bool
tty_grin_menu_command_imap4(const char *cmd) {
  tty_grin_imap4();
  return true;
}

bool
tty_grin_menu_command_compose(const char *cmd) {
  message m;
  
  prepare_message(m.root(), m.root());
  tty_grin_compose(m);
  return true;
}

bool
tty_grin_menu_command_addresses(const char *cmd) {
  tty_grin_addresses();
  return true;
}

bool
tty_grin_menu_command_resume(const char *cmd) {
  char buf[128];

  if(read_line(-1, 0, -1, _("Filename: "), buf, sizeof(buf), FILENAME)) {
    ifstream in(buf);
    
    if(in.bad()) report_error(strerror(errno));
    else {
      message m(in);
      in.close();
      tty_grin_compose(m);
    }
  }

  return true;
}

bool
tty_grin_menu_command_quit(const char *cmd) {
  return false;
}

bool
tty_grin_menu_command_help(const char *cmd) {
  static tty_grin_help_key_t help_keys[]={
    {"menu",      "inbox",     N_("Inbox")},
    {"menu",      "mail",      N_("Mail")},
    {"menu",      "news",      N_("News")},
    {"menu",      "imap4",     N_("IMAP4")},
    {"menu",      "compose",   N_("Compose")},
    {"menu",      "resume",    N_("Resume")},
    {"menu",      "addresses", N_("Addresses")},
    {"default",   "quit",      N_("Quit")},
    {"default",   "redraw",    N_("Redraw screen")},
    {"read-line", "cancel",    N_("Cancel")},
    {"read-line", "complete",  N_("Complete/search")},
    {0, 0, 0}
  };
  
  static tty_grin_help_t help={
    N_("Main menu"),
    N_("This is the help screen for the main menu.  You can receive help "
       "from within most screens by pressing ?.  Keys used in this screen "
       "are listed below.  Press a key to return to the menu."),
    help_keys
  };

  tty_grin_help(help);
  return true;
}

void
tty_grin_menu(void) {
  static tty_grin_menu_item_t items[]={
    {N_("INBOX"),        N_("Read new mail"),
     "menu",    "inbox", tty_grin_menu_command_inbox},
    {N_("MAIL"),         N_("Browse local mailboxes"),
     "menu",    "mail", tty_grin_menu_command_mail},
    {N_("NEWS"),         N_("Browse newsgroups"),
     "menu",    "news", tty_grin_menu_command_news},
    {N_("IMAP4"),        N_("Use IMAP4 profiles"),
     "menu",    "imap4", tty_grin_menu_command_imap4},
    {N_("COMPOSE"),      N_("Create a new message"),
     "menu",    "compose", tty_grin_menu_command_compose},
    {N_("RESUME"),       N_("Continue a postponed message"),
     "menu",    "resume", tty_grin_menu_command_resume},
    {N_("ADDRESSES"),    N_("Add, delete and modify addresses"),
     "menu",    "addresses", tty_grin_menu_command_addresses},
    {N_("QUIT"),         N_("Exit TTY-Grin"),
     "default", "quit", tty_grin_menu_command_quit},
    {N_("HELP"),         N_("How to use this program"),
     "default", "help", tty_grin_menu_command_help}
  };
  tty_grin_menu_t menu={sizeof(items)/sizeof(*items), 0, items};

  void *sigdata_old=sigdata_winch;
  SLSig_Fun_Type *f=SLsignal(SIGWINCH, tty_grin_menu_sigwinch);
  sigdata_winch=&menu;
  
  for(;;) {
    tty_grin_menu_draw(menu);
    SLsmg_refresh();
    const char *cmd=kbd_translate("menu", SLkp_getkey());

    if(cmd) {
      if(!strcasecmp(cmd, "up")) {
	if(!menu.selected--) menu.selected=menu.count-1;
      } else if(!strcasecmp(cmd, "down")) {
	if(++menu.selected>=menu.count) menu.selected=0;
      } else if(!strcasecmp(cmd, "jump-top")) {
	menu.selected=0;
      } else if(!strcasecmp(cmd, "jump-end")) {
	menu.selected=menu.count-1;
      } else if(!strcasecmp(cmd, "redraw")) {
	SLsmg_touch_lines(0, SLtt_Screen_Rows);
      } else if(!strcasecmp(cmd, "select")) {
	if(!menu.items[menu.selected].func
	   (menu.items[menu.selected].command)) {
	  sigdata_winch=sigdata_old;
	  SLsignal(SIGWINCH, f);
	  return;
	}
      } else {
	int i=0;
	
	while(i<menu.count&&strcasecmp(menu.items[i].command, cmd)) i++;
	
	if(i<menu.count) {
	  if(!menu.items[i].func(cmd)) {
	    sigdata_winch=sigdata_old;
	    SLsignal(SIGWINCH, f);
	    return;
	  }
	} else SLtt_beep();
      }
    }
  }
}
