// TTY-Grin
// Copyright (C) 2001-2003 Daniel Beer
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <fcntl.h>
#include <unistd.h>
#include <errno.h>
#include <fstream>
#include "grin.h"
#include "ctree.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

using namespace std;

class tty_grin_news_t : public ctree {
public:
  tty_grin_news_t(void);

protected:
  void draw(int, int, int, int);
  bool keypress(int);
  void winch(void);

private:
  void resize(void);

  bool command_post(const char *);
  bool command_delete(const char *);
  bool command_new_server(const char *);
  bool command_group(const char *);
  bool command_help(const char *);
  bool command_select(const char *);
};

tty_grin_news_t::tty_grin_news_t(void) : ctree(1) {
  int f=open(".", O_RDONLY, 0);

  if(!(chdir(getenv("HOME"))||chdir(".ttygrin"))) {
    DIR *d=opendir(".");
    if(d) {
      struct dirent *ent;
      while((ent=readdir(d))) {
	if(!strncmp(ent->d_name, "news:", 5)&&!chdir(ent->d_name)) {
	  ctree_node *server=new ctree_node(*this);

	  server->set_text(0, ent->d_name+5);

	  DIR *d=opendir(".");
	  if(d) {
	    while((ent=readdir(d))) if(*ent->d_name!='.') {
	      ctree_node *group=new ctree_node(*this, 0, server);
	      group->set_text(0, ent->d_name);
	    }
	    closedir(d);	    
	  }

	  chdir("..");
	}
      }
      closedir(d);
    }
  }

  fchdir(f);
  close(f);

  resize();
}

void
tty_grin_news_t::winch(void) {
  SLtt_get_screen_size();
  resize();
  ctree::winch();
}

void
tty_grin_news_t::resize(void) {
  set_width(0, SLtt_Screen_Cols);
}

void
tty_grin_news_t::draw(int r, int c, int rows, int cols) {
  SLsmg_set_color(SLC_STATUS);
  SLsmg_fill_region(r+rows-2, c, 1, cols, '-');
  SLsmg_gotorc(r+rows-2, c+4);
  SLsmg_printf(_(" News "));

  SLsmg_set_color(SLC_MAIN);
  SLsmg_fill_region(r+rows-1, c, 1, cols, ' ');

  ctree::draw(r, c, rows-2, cols);
}

bool
tty_grin_news_t::command_post(const char *cmd) {
  if(get_cursor()&&get_cursor()->get_parent()) {
    string group=get_cursor()->get_text(0);
    group+='@';
    group+=get_cursor()->get_parent()->get_text(0);
    
    message m;
    prepare_message(m.root(), m.root());
    m.root().headers()["Newsgroups"]=group;
    tty_grin_compose(m);
  }
  return true;
}

bool
tty_grin_news_t::command_delete(const char *cmd) {
  if(get_cursor()) {
    string name=getenv("HOME");
    name+="/.ttygrin/news:";
    if(get_cursor()->get_parent()) {
      if(yes_or_no(_("Delete this newsgroup (yes or no)? "))) {
	name+=get_cursor()->get_parent()->get_text(0);
	name+='/';
	name+=get_cursor()->get_text(0);
	if(unlink(name.c_str())) report_error(strerror(errno));
	else delete get_cursor();
      }
    } else {
      if(yes_or_no(_("Delete this server and "
		     "all newsgroups within (yes or no)? "))) {
	name+=get_cursor()->get_text(0);
	for(const ctree_node *n=get_cursor()->get_first();n;
	    n=n->get_next()) {
	  string group=name+'/'+n->get_text(0);
	  unlink(group.c_str());
	}
	if(rmdir(name.c_str())) report_error(strerror(errno));
	else delete get_cursor();
      }
    }
  }
  return true;
}

bool
tty_grin_news_t::command_new_server(const char *cmd) {
  char buf[128];

  if(read_line(-1, 0, -1, _("Hostname[:port]: "), buf, sizeof(buf))) {
    string name=getenv("HOME");
    name+="/.ttygrin/news:";
    name+=buf;
    if(mkdir(name.c_str(), 0700)) report_error(strerror(errno));
    else (new ctree_node(*this))->set_text(0, buf);
  }
  return true;
}

bool
tty_grin_news_t::command_group(const char *cmd) {
  if(get_cursor()) {
    ctree_node *p=get_cursor();
    if(p->get_parent()) p=p->get_parent();
    
    char buf[128];
    if(tolower(*cmd)=='n'?
       read_line(-1, 0, -1, _("Newsgroup: "), buf, sizeof(buf)):
       tty_grin_groupbrowse(p->get_text(0), buf, sizeof(buf))) {
      string name=getenv("HOME");
      name+="/.ttygrin/news:";
      name+=p->get_text(0);
      name+='/';
      name+=buf;
      
      int f=open(name.c_str(), O_CREAT|O_WRONLY, 0600);
      if(f<0) report_error(strerror(errno));
      else {
	(new ctree_node(*this, 0, p))->set_text(0, buf);
	close(f);
      }
    }
  }
  return true;
}

bool
tty_grin_news_t::command_help(const char *cmd) {
  static tty_grin_help_key_t help_keys[]={
    {"news",    "delete",       N_("Delete")},
    {"news",    "new-server",   N_("Create server")},
    {"news",    "new-group",    N_("Create group")},
    {"news",    "pick-group",   N_("Browse for groups")},
    {"news",    "post",         N_("Post to")},
    {"tree",    "collapse",     N_("Collapse")},
    {"tree",    "expand",       N_("Expand")},
    {"tree",    "collapse-all", N_("Collapse all")},
    {"tree",    "expand-all",   N_("Expand all")},
    {"default", "quit",         N_("Back")},
    {0, 0, 0}
  };

  static tty_grin_help_t help={
    N_("News"),
    N_("Create and manage subscriptions to newsgroups on multiple news "
       "servers accessable via NNTP."),
    help_keys
  };

  tty_grin_help(help);
  return true;
}

bool
tty_grin_news_t::command_select(const char *cmd) {
  if(get_cursor()&&get_cursor()->get_parent()) {
    newsgroup n(get_cursor()->get_parent()->get_text(0),
		get_cursor()->get_text(0));
    tty_grin_newsgroup(n);
  } else SLtt_beep();
  return true;
}

bool
tty_grin_news_t::keypress(int key) {
  const char *cmd=kbd_translate("news", key);

  static struct {
    const char *command;
    bool (tty_grin_news_t::*func)(const char *);
  } commands[]={
    {"post",       &tty_grin_news_t::command_post},
    {"delete",     &tty_grin_news_t::command_delete},
    {"new-server", &tty_grin_news_t::command_new_server},
    {"new-group",  &tty_grin_news_t::command_group},
    {"pick-group", &tty_grin_news_t::command_group},
    {"help",       &tty_grin_news_t::command_help},
    {"select",     &tty_grin_news_t::command_select},
    {0, 0}
  };

  if(cmd) {
    int i=0;
    while(commands[i].command&&
	  strcasecmp(commands[i].command, cmd)) i++;
    
    if(commands[i].command) return (this->*commands[i].func)(cmd);
  }

  return ctree::keypress(key);
}

void
tty_grin_news(void) {
  tty_grin_news_t n;

  n.set_cursor_color(SLC_CURSOR);
  n.run();
}
