// TTY-Grin
// Copyright (C) 2001-2003 Daniel Beer
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#define _GNU_SOURCE
#include <fstream>
#include <time.h>
#include <errno.h>
#include "grin.h"
#include "ctree.h"
#include "tempfile.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

using namespace std;

class tty_grin_newsgroup_t : public ctree {
public:
  tty_grin_newsgroup_t(newsgroup&);

protected:
  virtual void winch(void);
  virtual bool keypress(int);
  virtual void draw(int, int, int, int);

private:
  void resize(void);
  int flagify(ctree_node *);
  void reflagify(ctree_node *);
  char flagmap(int);

  newsgroup *ng;

  bool command_toggle(const char *);
  bool command_mark_all_read(const char *);
  bool command_post(const char *);
  bool command_select(const char *);
  bool command_help(const char *);
  bool command_quit(const char *);
};

void
tty_grin_newsgroup_t::reflagify(ctree_node *n) {
  while(n->get_parent()) n=n->get_parent();
  flagify(n);
}

bool
tty_grin_newsgroup_t::command_toggle(const char *cmd) {
  int flag=0;

  switch(tolower(cmd[7])) {
  case 'r': flag=message_descriptor::READ; break;
  case 'm': flag=message_descriptor::MARKED; break;
  case 'd': flag=message_descriptor::DELETED; break;
  }

  if(get_cursor())
    ((message_descriptor *)get_cursor()->get_data())->
      toggle_flag(flag);
  reflagify(get_cursor());
  if(get_cursor()->down()) set_cursor(get_cursor()->down());

  return true;
}

bool
tty_grin_newsgroup_t::command_select(const char *cmd) {
  if(get_cursor()) {
    newsgroup_message *m=(newsgroup_message *)get_cursor()->get_data();
    tempfile t;
    
    ofstream out(t.filename());
    if(out.good()) {
      m->extract(out);
      out.close();
      ifstream in(t.filename());
      if(in.good()) {
	message msg(in);
	in.close();
	m->set_flag(message_descriptor::READ);
	tty_grin_message(msg, m);
	reflagify(get_cursor());
      }
    }      
  }
  return true;
}

bool
tty_grin_newsgroup_t::command_mark_all_read(const char *cmd) {
  if(yes_or_no(_("Mark all messages as read (yes or no)? "))) {
    for(int i=0;i<ng->size();i++)
      (*ng)[i].set_flag(message_descriptor::READ);
    for(ctree_node *n=get_first();n;n=n->get_next()) flagify(n);
  }
  return true;
}

bool
tty_grin_newsgroup_t::command_post(const char *cmd) {
  message m;
  
  prepare_message(m.root(), m.root());
  string group=ng->get_group_name();
  group+='@';
  group+=ng->get_server_name();
  m.root().headers()["Newsgroups"]=group;
  tty_grin_compose(m);

  return true;
}

bool
tty_grin_newsgroup_t::command_help(const char *cmd) {
  static tty_grin_help_key_t help_keys[]={
    {"newsgroup", "toggle-read",    N_("Toggle read")},
    {"newsgroup", "toggle-marked",  N_("Toggle special")},
    {"newsgroup", "toggle-deleted", N_("Toggle deleted")},
    {"tree",      "collapse",       N_("Collapse")},
    {"tree",      "expand",         N_("Expand")},
    {"tree",      "collapse-all",   N_("Collapse all")},
    {"tree",      "expand-all",     N_("Expand all")},
    {"default",   "find-forward",   N_("Search forward")},
    {"default",   "find-back",      N_("Search back")},
    {"newsgroup", "post",           N_("Post to group")},
    {"newsgroup", "mark-all-read",  N_("Mark all as read")},
    {"default",   "quit",           N_("Back")},
    {0, 0, 0}
  };

  static tty_grin_help_t help={
    N_("Newsgroup"),
    N_("Read news messages from servers accessable via NNTP.  Message "
       "markings appear as follows:\n\n"
       "  R - New/unread\n"
       "  A - Answered\n"
       "  * - Special\n"
       "  D - Deleted"),
    help_keys
  };
  
  tty_grin_help(help);
  return true;
}

bool
tty_grin_newsgroup_t::command_quit(const char *cmd) {
  ctree_node *n=get_first();
  while(n&&!(((message_descriptor *)n->get_data())->get_flags()&
	     message_descriptor::DELETED))
    n=n->down(true);
  if(n&&yes_or_no(_("Expunge deleted messages (yes or no)? ")))
    ng->expunge();
  return false;
}

bool
tty_grin_newsgroup_t::keypress(int key) {
  const char *cmd=kbd_translate("newsgroup", key);

  static struct {
    const char *command;
    bool (tty_grin_newsgroup_t::*func)(const char *);
  } commands[]={
    {"toggle-read",    &tty_grin_newsgroup_t::command_toggle},
    {"toggle-marked",  &tty_grin_newsgroup_t::command_toggle},
    {"toggle-deleted", &tty_grin_newsgroup_t::command_toggle},
    {"mark-all-read",  &tty_grin_newsgroup_t::command_mark_all_read},
    {"post",           &tty_grin_newsgroup_t::command_post},
    {"select",         &tty_grin_newsgroup_t::command_select},
    {"help",           &tty_grin_newsgroup_t::command_help},
    {"quit",           &tty_grin_newsgroup_t::command_quit},
    {0, 0}
  };

  if(cmd) {
    int i=0;
    while(commands[i].command&&
          strcasecmp(commands[i].command, cmd)) i++;
    
    if(commands[i].command) return (this->*commands[i].func)(cmd);
  }

  return ctree::keypress(key);
}

void
tty_grin_newsgroup_t::draw(int r, int c, int rows, int columns) {
  SLsmg_set_color(SLC_STATUS);
  SLsmg_fill_region(r+rows-2, c, 1, columns, '-');
  SLsmg_gotorc(r+rows-2, c+4);
  SLsmg_printf(_(" %s on %s -- %d articles "),
	       ng->get_group_name(), ng->get_server_name(),
	       ng->size());

  SLsmg_set_color(SLC_MAIN);
  SLsmg_fill_region(r+rows-1, c, 1, columns, ' ');

  ctree::draw(r, c, rows-2, columns);
}

char
tty_grin_newsgroup_t::flagmap(int x) {
  if(x&message_descriptor::DELETED) return 'D';
  if(x&message_descriptor::MARKED) return '*';
  if(x&message_descriptor::ANSWERED) return 'A';
  if(x&message_descriptor::READ) return 0;
  return 'N';
}

int
tty_grin_newsgroup_t::flagify(ctree_node *node) {
  newsgroup_message *msg=(newsgroup_message *)node->get_data();
  int flags=0;

  for(ctree_node *n=node->get_first();n;n=n->get_next())
    flags|=flagify(n)^message_descriptor::READ;

  char buf[5], one, two;

  one=flagmap(msg->get_flags());
  two=flagmap(flags^message_descriptor::READ);

  if(one) {
    if(two) snprintf(buf, sizeof(buf), "%c(%c)", one, two);
    else {
      buf[0]=one;
      buf[1]=0;
    }
  } else {
    if(two) snprintf(buf, sizeof(buf), "(%c)", two);
    else buf[0]=0;
  }

  node->set_text(0, buf);

  flags|=msg->get_flags()^message_descriptor::READ;
  flags^=message_descriptor::READ;

  return flags;
}

void
tty_grin_newsgroup_t::resize(void) {
  set_width(0, 5);
  set_width(1, (SLtt_Screen_Cols-25)/2);
  set_width(2, (SLtt_Screen_Cols-25)/2);
  set_width(3, SLtt_Screen_Cols);
}

void
tty_grin_newsgroup_t::winch(void) {
  SLtt_get_screen_size();
  resize();
  ctree::winch();
}

tty_grin_newsgroup_t::tty_grin_newsgroup_t(newsgroup& n) : ctree(4) {
  ng=&n;

  map<const newsgroup_message *, ctree_node *> threadme;

  const char *kfcmd=config_get("kill-filter");
  if(kfcmd) {
    tempfile tin, tout;
    ofstream out(tout.filename());
    if(out.good()) {
      for(int i=0;i<n.size();i++)
	out<<i<<'\t'<<n[i].get_subject()<<'\t'<<n[i].get_from()<<'\t'<<
	  n[i].get_date()<<endl;
      out.close();

      // This would probably be faster done through a pipe.
      char buf[1024];
      snprintf(buf, sizeof(buf), "%s news %s %s < %s > %s",
	       kfcmd, n.get_server_name(), n.get_group_name(),
	       tout.filename(), tin.filename());
      system(buf);

      ifstream in(tin.filename());
      if(in.good()) {
	in.getline(buf, sizeof(buf));
	while(!(in.eof()||in.fail())) {
	  int num=atoi(buf);
	  if(num>=0&&num<n.size())
	    n[num].set_flag(message_descriptor::USR);
	  in.getline(buf, sizeof(buf));
	}
	in.close();
      } else report_error(strerror(errno));
    } else report_error(strerror(errno));
  }

  for(int i=0;i<n.size();i++)
    if(!(n[i].get_flags()&message_descriptor::USR||
	 config_should_kill(n[i].get_from()))) {
      ctree_node *m=new ctree_node(*this);

      m->set_text(1, n[i].get_subject());
      m->set_text(2, n[i].get_from());
      
      const char *rawdate=n[i].get_date();
      if(strlen(rawdate)>5) {
	if(rawdate[3]==',') rawdate+=5;
	struct tm date;
	strptime(rawdate, "%d %b %Y %T", &date);
	char buf[64];
	strftime(buf, sizeof(buf), "%F", &date);
	buf[sizeof(buf)-1]=0;
	m->set_text(3, buf);
      }
      
      m->set_data((void *)&n[i]);
      threadme.insert(pair<const newsgroup_message *,
		      ctree_node *>(&n[i], m));
    }

  for(int i=0;i<n.size();i++) n[i].clear_flag(message_descriptor::USR);

  for(ctree_node *n=get_first();n;) {
    ctree_node *nn=n->get_next();

    newsgroup_message *m=(newsgroup_message *)n->get_data();
    newsgroup_message *mp=m->get_parent();
    if(mp) {
      map<const newsgroup_message *, ctree_node *>::
	const_iterator i=threadme.find(mp);
      if(i!=threadme.end()) n->move(0, i->second);
    }

    n=nn;
  }

  for(ctree_node *n=get_first();n;n=n->get_next()) flagify(n);

  resize();
}

void
tty_grin_newsgroup(newsgroup& n) {
  tty_grin_newsgroup_t g(n);

  g.set_cursor_color(SLC_CURSOR);

  g.run();
}
